/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <stdarg.h>
#include <ctype.h>
#include <dirent.h>

#include <iostream>
#include <fstream>
#include <vector>
#include <algorithm>  

#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif

#include "text.hpp"
#include "intarr.hpp"
#include "textarr.hpp"
#include "array64.hpp"
#include "dyn.hpp"
#include "array.hpp"


text sol_pair::toText() {
  text t = key;
  t.cat(" : ");
  t.cat(value->toText().c_str());
  return t;
}

char *sol_pair::toStr() {
  return value->toStr();
}


/** \file array.cpp
  File of the associative, Php compatible, array and dict.
  */


extern void errorMsg(char *);
int array::nomatch()   { errorMsg("array type mismatch"); return 0; }

dyn array::nullelement = dyn(false);

/* Constructors */

array::array()
{
  freeKey = 0;
  indexer = 0;
}

 /**
   Build an associative array from an array of strings of chars.
   Indexes are generated from 0 to n.
   */

array::array(char *a[]) { set(a); }

array::array(dyn x) {
    freeKey = 0;
    indexer = 0;
    push(x);
}

array::array(const array &a) {
  keyList.assign(a.keyList.begin(), a.keyList.end());
  data.assign(a.data.begin(), a.data.end());
  freeKey = a.freeKey;
  indexer = 0;
}

array::array(array *a) {
  keyList.assign(a->keyList.begin(), a->keyList.end());
  data.assign(a->data.begin(), a->data.end());
  freeKey = a->freeKey;
  indexer = 0;
}

// create an array from a type array
array::array(_intArr b) { set(b); }
array::array(_arr32 b) { set(b); }
array::array(_array64 b) { set(b); }

array::~array()
{
  keyList.clear();
  data.clear();
}


void array::clear()
{
  keyList.clear();
  data.clear();
  indexer = 0;
  freeKey = 0;
}

void array::pushKey(const text &s)
{
  if(text::isNumber(s.data))
    keyList.push_back(text(freeKey++));
  else
    keyList.push_back(s);
}

void array::set(const array &a)  {
  keyList.assign(a.keyList.begin(), a.keyList.end());
  data.assign(a.data.begin(), a.data.end());
  freeKey = a.freeKey;
  indexer = 0;
}
void array::set(array *a) {
  keyList.assign(a->keyList.begin(), a->keyList.end());
  data.assign(a->data.begin(), a->data.end());
  freeKey = a->freeKey;
  indexer = 0;
}

void array::set(_intArr b) {
  clear();
  if(&b == NULL) return;
  for(int i = 0; i < b.size(); i++) {
    push(dyn(b.at(i)));
  }
}

void array::set(_arr32 b) {
  clear();
  if(&b == NULL) return;
  for(int i = 0; i < b.size(); i++) {
    push(dyn(b.at(i)));
  }
}


void array::set(_array64 b) {
  clear();
  if(&b == NULL) return;
  for(int i = 0; i < b.size(); i++) {
    push(dyn(b.at(i)));
  }
}



void array::set(dyn d) {
  if(d.flag != TYPE_ARRAY) { clear(); push(d); return; }
  set(d.a);
}

void array::set(char *x[])
{
  int i = 0;
  this->clear();
  while(x[i] != NULL)  push(dyn(x[i++]));
}

dyn array::get(int n)
{
  char buffer[64];
  sprintf(buffer, "%d", n);
  return get(buffer);
}

dyn array::get(const text &a)
 {
  for (unsigned int i = 0; i < keyList.size(); i++)
   {
     if(strcmp(a.data, keyList[i].data) == 0) return data[i];
   }
  return dyn(false);
 }

dyn array::get(const char *a)
{
  for (unsigned int i = 0; i < keyList.size(); i++)
   {
     if(strcmp(a, keyList[i].data) == 0) return data[i];
   }
  return dyn(false);
}



void array::add(const array &a)
{
  const array *x = &a;
  add((array *) x);
}


void array::add(array *x)
{
  int k;
  for(int i = 0; i < x->size(); i++)
  {
    k = hasStrKey(x->keyList[i]);
    if(k != -1) data[k].set(x->data[i]);
    else
    {
      pushKey(x->keyList[i]);
      data.push_back(x->data[i]);
    }
  }
  pack();
}


void array::clone(array *y)
 {
   y->keyList.assign(keyList.begin(), keyList.end());
   y->data.assign(data.begin(), data.end());
   y->freeKey = freeKey;
   y->indexer = 0;
 }

//===== OPERATORS

void array::operator=(const array &a)
{
  keyList.assign(a.keyList.begin(), a.keyList.end());
  data.assign(a.data.begin(), a.data.end());
  freeKey = a.freeKey;
  indexer = 0;
  return;
}

void array::operator=(array *a)     { set(a); }
void array::operator=(dyn a)        {  set(a); }
void array::operator=(_intArr &a)   { set(a); }
void array::operator=(_arr32 &a)    { set(a); }
void array::operator=(_array64 &a)  { set(a); }


 array array::operator+(array *x)
 {
   array y = array(this);
   y.add(x);
   return y;
 }

 array array::operator+(dyn x)
 {
   array y = array(this);
   y.push(x);
   return y;
 }



//array array::operator-(const array &b) {  return sub(b); }
array array::operator-(array *b)
{
   dyn v;
   int i,j;
   int l = b->size();
   array y = array();
   text k;

   for(i = 0; i < size(); i++)
   {
     k = keyList[i];
     v.set(data[i]);
     for(j = 0; j  < l; j++)
     {
       if(v == &b->data[j]) goto skip;
     }
     if(text::isNumber(k.data))  y.push(v);
     else y.push(k, v);
     skip:;
   }
   return y;
};

array array::operator-(dyn x)
{
   array y = array();
   for(int i = 0; i < size(); i++)
   {
     if(data[i] != x)
     {
       y.push(keyList[i], data[i]);
     }
   }
   return y;
}

// array array::operator&(array a) { return operator&(a); }
 array array::operator&(array *b)
 {
   dyn v;
   int i,j;
   int l = b->size();
   array *y = new array();
   text k;

   for(i = 0; i < size(); i++)
   {
     k = keyList[i];
     v.set(data[i]);
     for(j = 0; j  < l; j++)
     {
       if(v == b->data[j])
       {
         if(text::isNumber(k.data))
           y->push(v);
         else
         {
           y->keyList.push_back(k);
           y->data.push_back(v);
         }
       }
     }
   }
   y->pack();
   return *y;
 };


//array array::operator|(array a){ return operator|(a); };

array array::operator|(array *b)
{
  unsigned int i, l = size();
  array *y = new array();

  for(i = 0; i < l; i++)
  {
    y->pushUnic(keyList[i], data[i]);
  }
  y->add(b);
  y->pack();

  return y->unique();
}

//array array::operator^(array a){ return operator|(a); };
array array::operator^(array *b)
{
  int i;
  array *y = new array();

  for(i = 0; i < size(); i++)
  {
    if(b->findValue(data[i]) == -1)
      y->pushUnic(keyList[i], data[i]);
  }
  for(i = 0; i < b->size(); i++)
  {
    if(findValue(b->data[i]) == -1)
      y->pushUnic(b->keyList[i], b->data[i]);
  }
  return *y;
}

/*void array::operator+=(const array &x) { this->add(x); return *this; }*/
array &array::operator+=(array *x)
{
   this->add(x);
   return *this;
}
array &array::operator+=(dyn x)
{
   //puts("by array");
   push(x);
   return *this;
}


// void array::operator-=(array a) { operator-=(a); }
 void array::operator-=(array *x)
 {
   array y = array(this);
   set(y.operator-(x));
 }

 void array::operator-=(dyn x) {
   //if(x.flag ==  TYPE_ARRAY) operator-=(x.a);
   array y = array(this);
   set(y - x);
 }

// void array::operator&=(array b) { operator&=(b); }
 void array::operator&=(array *b)
 {
   array y = operator&(b);
   set(y);
 }
// void array::operator|=(array b) { operator|=(b); }
 void array::operator|=(array *b)
 {
   array y = operator|(b);
   set(y);
 }
// void array::operator^=(array b) { operator^=(b); }
 void array::operator^=(array *b)
 {
   array y = operator^(b);
   set(y);
 }




bool array::operator==(array *b)
 {
   int l = size();
   if(b->size() != l) return false;
   for(int i = 0; i < l; i++)
   {
     if(strcmp(keyList[i].data, b->keyList[i].data) != 0) return false;
     if(data[i] != b->data[i]) return false;
   }
   return true;
 }

// bool  array::operator==(array b) { return operator==(b); };
 bool  array::operator==(bool x)  { return (size() != 0) == x; }

// bool  array::operator!=(array a)  { return !operator==(a); };
 bool  array::operator!=(array *a) { return !operator==(a); };
 bool  array::operator!=(bool x) { return !operator==(x); }

// bool array::operator<(array a) { return operator<(a); };
 bool array::operator<(array *a)
 {
   int l = a->size();
   if(size() < l) return true;
   if(size() > l) return false;
   for(int i = 0; i < l; i++)
   {
     if(data[i] > a->data[i]) return false;
     if(data[i] < a->data[i]) return true;
   }
   return false;
 }

// bool array::operator>(array a) { return operator>(a); };
 bool array::operator>(array *a)
 {
   int l = a->size();
   if(size() < l) return false;
   if(size() > l) return true;
   for(int i = 0; i < l; i++)
   {
     if(data[i] > a->data[i]) return true;
     if(data[i] < a->data[i]) return false;
   }
   return false;
 }

bool array::operator!() { return size() == 0; }

array &array::operator,(dyn x) { push(x); return *this; }

//===== util

 bool array::in(const dyn &d)  {
  for(unsigned int i = 0; i < data.size(); i++)  {
   if(data[i] == d) return true;
  }
  return false;
 }

 bool array::in(dyn *d)  {
    for(unsigned int i = 0; i < data.size(); i++)  {
        if(data[i] == d) return true;
    }
    return false;
 }

/* is the textual key already in dict? */
signed int array::hasStrKey(const text &s) {
    if(!text::isNumber(s.data))  {
    for(unsigned i = 0; i < keyList.size(); i++)
        if(strcmp(s.data, keyList[i].data) == 0) return i;
    }
    return -1;
}

signed int array::findKey(const text &s) {
    for(unsigned i = 0; i < keyList.size(); i++)   {
    if(strcmp(s.data, keyList[i].data) == 0) return i;
    }
    return -1;
}

signed int array::findKey(const char *s)
{
  int l = size();
  for(int i = 0; i < l; i++)
  {
   if(strcmp(s, keyList[i].data) == 0) return i;
  }
  return -1;
}


signed int array::findKey(int n) { return findKey(text(n)); }

signed int array::findValue(dyn *d) { return findValue(*d); }
signed int array::findValue(const dyn &d)
{
  for(int i = 0; i < size(); i++)
  {
   if(data[i] == d) return i;
  }
  return -1;
}

/*
  Find last value
*/

signed int array::findLastValue(dyn *d) { return findLastValue(*d); }
signed int array::findLastValue(const dyn &d)
{
  for(int i = size() - 1; i >= 0; i--)
  {
   if(data[i] == d) return i;
  }
  return -1;
}

 //===== iterator

 void array::reset() { indexer = 0; }
 
 bool array::outOfBound() { 
     return indexer >= size(); 
 }

 bool array::bound()  { 
     return(++indexer < size()); 
 }
 
 bool array::inBound() {
  if(indexer < size()) { 
      return true;
  }
  return false;
 }
 
 dyn array::inc()  {
  if(++indexer < size()) { 
      return data[indexer]; 
  }
  indexer = size();
  return dyn(false);
 }

 dyn array::dec() {
  if(--indexer >= 0) return data[indexer];
  indexer = 0;
  return dyn(false);
 }

 dyn array::current() { return value(indexer); }

 dyn array::value() { return value(indexer); }

 dyn array::value(int i) {
   //std::cout << i << "/" << size() << " " << data[i].toStr() << std::endl;
   if(i >=0 && i < size())  return data[i];
   return dyn(false);
  }

 text array::key() {
   if(indexer >= 0 && indexer < (int) keyList.size())  return keyList[indexer];
   return text();
  }
 text array::key(int i)    {
   if(i >= 0 && i < (int) keyList.size())  return keyList[i];
   return text();
  }
 int array::index()
 {
   if(indexer >=0 && indexer < (int) keyList.size()) return atoi(keyList[indexer].c_str());
   return -1;
 }

dyn array::begin()
 {
   indexer = 0;
   if(data.size() > 0) return data.front();
   return dyn(false);
}
dyn array::end()   {
   indexer = 1;
   if(data.size() > 0)
   {
     indexer = (int) data.size() -1;
     return data.back();
   }
   return dyn(false);
}

//===== stack

void array::push(const dyn &a)
 {
   //dyn *xxx = new dyn(a);
   //std::cout << "push f=" << xxx->flag << " " << xxx->toStr() << std::endl;
   if(a.flag == TYPE_PAIR) push(a.p);
   else
   {
     keyList.push_back(text(freeKey++));
     data.push_back(a);
     //std::cout << "back: " << data.back() << std::endl;
   }
 }

 void array::push(const char *s)
 {
	keyList.push_back(text(freeKey++));
	dyn *d = new dyn(s);
	//printf("BACK %s\n", d.toStr());
	//printf("data %lu \n", data);
	data.push_back(*d);
	//puts("PUSHED");
 }

 void array::push(dyn *a)
 {
   if(a->flag == TYPE_PAIR) push(a->p->key, a->p->value);
   else
   {
     keyList.push_back(text(freeKey++));
     data.push_back(*a);
   }
 }

 void array::push(sol_pair *p)
 {
    push(p->key, p->value);
    return;
 }

 void array::push(const text &k, dyn v)
 {
    int i;
    if((i = findKey(k)) != -1)
     {
      data[i].set(v);
      return;
     }
    keyList.push_back(k);
    data.push_back(v);
 }

 void array::pushAdd(const text &k, dyn v)
 {
    keyList.push_back(k);
    data.push_back(v);
 }

 void array::pushAdd(int k, dyn v)
 {
    if(k < freeKey) k = freeKey;
    keyList.push_back(text(k));
    data.push_back(v);
    freeKey = k+1;
 }


 void array::pushUnic(const text &k, dyn v)
 {
     if(findValue(v) == -1)
     {
       if(text::isNumber(k.data))
        {
         keyList.push_back(text(freeKey++));
        }
       else
         keyList.push_back(k);
       data.push_back(v);
       return;
     }
 }

 dyn array::pop()
 {
   dyn x = dyn(false);
   indexer = 0;
   if(!empty())
   {
     x.set(data.back());
     keyList.pop_back();
     data.pop_back();
   }
   return x;
 }

dyn array::shift()
{
	if(empty())  return dyn(false);
	else
	{
		dyn *x = &data[0];
		erase(0);
		return *x;
	}
}

void array::unshift(dyn a)
{
  keyList.insert(keyList.begin(), text("0"));
  data.insert(data.begin(), a);
  pack();
}

void array::insert(int n, dyn d)
{
  if(n < 0) n = size() + n;
  if(n < 0) return;
  if(n >= size())
  {
    if(d.flag == TYPE_ARRAY) add(d.a);
    else  push(d);
    return;
  }

  if(d.flag == TYPE_ARRAY)
  {
   keyList.insert(keyList.begin() + n, d.a->keyList.begin(), d.a->keyList.end());
   data.insert(data.begin() + n, d.a->data.begin(), d.a->data.end());
  }
  else
  {
   keyList.insert(keyList.begin() + n, text(65000));
   data.insert(data.begin() + n, d);
  }
  pack();
}

//===== RANGE

 void array::pack()
 {
   freeKey = 0;
   for(int i = 0; i < size(); i++)
   {
     //printf("pack: %d) %s %s\n", i, keyList[i].data, data[i].toStr());
     if(text::isNumber(keyList[i].data))
     {
       keyList[i] = text(freeKey++);
     }
   }
 }

 // erase from first to last included
 void array::erase(int first, int last)
 {
    keyList.erase(keyList.begin() + first, keyList.begin() + last + 1);
    data.erase(data.begin() + first, data.begin() + last + 1);
    pack();
 }

 void array::erase(int pos)
 {
    //std::cout << "erase " << pos << std::endl;
    //display();
    keyList.erase(keyList.begin() + pos);
    data.erase(data.begin() + pos);
    //display();
    pack();
 }


 array array::slice(int f, int l, int incl)
 {
   array x;

   if(empty()) goto bye;
   l += incl;
   l--;
   if(l < 0) l += size();
   if(l < 0) goto bye;
   if(l >= size()) l = size() - 1;

   if(f < 0) f += size();
   if(f < 0) f = 0;
   if(l < f) goto bye;
   if(f >= size()) goto bye;

   x.keyList.assign(keyList.begin() + f, keyList.begin() + l + 1);
   x.data.assign(data.begin() + f, data.begin() + l + 1);
   x.pack();
bye:
   return x;
 }

 void array::splice(int b, int e)
 {
   if(b < 0) b = size() + b;
   if(b < 0) b = 0;
   if(e < 0) e = size() + e;
   if(e < 0) return;
   if(e >= size()) e = size() - 1;
   if(e < b) return;
   erase(b, e);
 }

 void array::splice(int b, int e, int inc)
 {
   splice(b, e + inc - 1);
 }


 void array::splice(int f)  { erase(f); }

 void array::splice(int i, int last, array a, int incl) { splice(i, last, &a, incl); }
 void array::splice(int i, int last, array *a, int incl)
 {
  if(empty()) { operator=(a); return; }
  if(last < 0) last = size() - last;
  if(i < 0) i += size();
  if(i < 0) i = 0;
  last += incl;
  last--;
  if(last >= size()) last = size() - 1;
  //cout << i << " " << last << " " << size() << std::endl;
  if(i >= size()) { add(a); return; }
  if(last < i) return;
  if(last < 0) return;
  if((last + 1) == i)
  {
    if(!a->empty()) insert(i, dyn(a));
    return;
  }
  if(a->empty()) { erase(i, last); return; }

  int j = 0;
  while((i <= last) && (j < a->size()))
  {
    keyList[i].set(a->keyList[j]);
    data[i].set(a->data[j]);
    i++;
    j++;
  }
  if(last >= i) erase(i, last);
  if(j < a->size())
  {
   keyList.insert(keyList.begin() + i, a->keyList.begin() + j, a->keyList.end());
   data.insert(data.begin() + i, a->data.begin() + j, a->data.end());
  }
  pack();
 }


 void array::splice(int i, int last, dyn a, int incl)
 {
  if(i < 0) i = size() - i;
  if(i < 0) i = 0;
  if(last < 0) last = size() - last;
  last += incl;
  last --;
  if(last >= size()) last = size() - 1;
  if(last < i) return;
  if(last < 0) return;

  if(i >= size()) { push(a); return; }
  if(last == i)   { insert(i, a); }

  keyList[i].set(text("0"));
  data[i].set(a);
  erase(i + 1, last);
 }

//===== TRANSFORM

/*
  if data is a not a number, search if exists further
  if yes, wait for the next occurence to push
  if not a number push it providing not already pushed
*/

array array::unique()
{
  int i, l = size();
  array y = array();

  for(i = 0; i < l; i++)
  {
    if(y.findValue(data[i]) == -1)
      y.push(keyList[i], data[i]);
  }
  return y;
}


void array::swapData(int a, int b)
{
  dyn x;
  x.set(data[a]);
  data[a].set(data[b]);
  data[b].set(x);
}

void array::swapElem(int a, int b)
{
  text s = keyList[a];
  dyn x;
  x.set(data[a]);
  keyList[a].set(keyList[b]);
  data[a].set(data[b]);
  keyList[b].set(s);
  data[b].set(x);
}

void array::keysSort(int first,int last)
{
 int f=first;
 int l=last;
 dyn item;
 item.set(keyList[(f + l) / 2]);

 do
 {
  while(keyList[f] < item) f++;
  while(item < keyList[l]) l--;
  if(f <= l)
  {
   swapElem(f, l);
   f++;
   l--;
  }
 } while(f <= l);

 if(first < l)  keysSort(first, l);
 if(f < last)   keysSort(f, last);
 return;
}

void array::dynSort(int first,int last)
{
 int f=first;
 int l=last;
 dyn item;
 item.set(data[(f + l) / 2]);

 do
 {
  while(data[f] < item) f++;
  while(item < data[l]) l--;
  if(f <= l)
  {
   swapElem(f, l);
   f++;
   l--;
  }
 } while(f <= l);

 if(first < l)  dynSort(first, l);
 if(f < last)   dynSort(f, last);
 return;
}

void array::dynRSort(int first, int last)
{
 dynSort(first, last);
 reverse();
 return;
}


void array::sort(int x)  { x=0; dynSort(0, size() - 1); pack(); }
void array::rSort(int x) { x=0; dynRSort(0, size()-1); }
void array::kSort(int x) { x=0; keysSort(0, size() - 1); }

array array::keys()
{
  array x = array();
  for(unsigned int i = 0; i < keyList.size(); i++) x.push(dyn(keyList[i]));
  return x;
}

array array::values()
{
  array x = array();
  x.data.assign(data.begin(), data.end());
  for(unsigned int i = 0; i < data.size(); i++)
   x.keyList.push_back(text(i));
  return x;
}

dyn array::min()
{
  double m = DBL_MAX;
  dyn d;
  double x = 0;
  for(int i = 0; i < size(); i++)
  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) x = d.toReal();
    else
     if(d.flag == TYPE_TEXT)  x = d.t->toNumber();
    if(x < m) m = x;
  }
  return dyn(m);
}

dyn array::max()
{
  double m = 0;
  dyn d;
  double x = 0;
  for(int i = 0; i < size(); i++)
  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) x = d.toReal();
    else
     if(d.flag == TYPE_TEXT)  x = d.t->toNumber();
    if(x > m) m = x;
  }
  return dyn(m);
}

/*
  Functional map method
  */

array array::map(bool (*f)(dyn))
{
  int s = data.size();
  array temp;
  for(int i = 0; i < s; i++)
  {
    if(f(data[i]))
      temp.push(data[i]);
  }
  return temp;
}

double array::sum()
{
  double total = 0;
  dyn d;
  for(int i = 0; i < size(); i++)
  {
    d.set(data[i]);
    if(d.flag == TYPE_NUMBER) total += d.toReal();
    else
     if(d.flag == TYPE_TEXT)
     {
       total += d.t->toNumber();
     }
  }
  return total;
}

//===== OPERATIONS

text array::join(const char *sep)
{
  int lsep, si, sum, offset;
  int i;

  lsep = strlen(sep);
  si = size();
  sum = 0;
  offset = 0;

  for(i = 0; i < si; i++)
  {
    if(data[i].flag == TYPE_TEXT)
    {
     sum+= data[i].t->length();
     sum+= lsep;
    }
  }

  char *nstr = (char *) new char[sum+1];
  for(i = 0; i < si; i++)
  {
   if(data[i].flag == TYPE_TEXT)
   {
     if(offset > 0)
     {
      strcpy(nstr + offset, sep);
      offset += lsep;
     }
     strcpy(nstr + offset, data[i].t->data);
     offset += data[i].t->length();
     nstr[offset] = '\0';
   }
  }
  nstr[offset] = '\0';
  return text(nstr);
}

array array::reverse()
{
  array y = array();
  for(int i = size() - 1; i >= 0; i--)
  {
    y.keyList.push_back(keyList[size() - i - 1]);
    y.data.push_back(data[i]);
  }
  return y;
}


std::ostream& operator<< (std::ostream& out, array *d)
{
  return out << d->toText().data;
}


void array::display() { display(0); }
void array::display(int tab)
{
 int l = size();
 dyn x;
 std::cout << "Array(\n";
 for(int i = 0; i < l; i++)
 {
   //std::cout << "flag=" << data[i].flag << " " << data[i].toStr() << std::endl;
   for(int j = 0; j < tab; j++) std::cout << "     ";
   std::cout << "   " << keyList[i].data;
   std::cout << " : ";
   x.set(data[i]);
   if(x.flag == TYPE_ARRAY) x.a->display(tab + 1);
   else
    std::cout << x.toStr() << std::endl;
 }
 for(int j = 0; j < tab; j++) std::cout << "     ";
 std::cout << ")\n";
}

// Convert array to string

char *array::toString(char *s) {
	int l = size();
	text *t = new text();
	for(int i = 0; i < l; i++) 	{
		if(i>0) t->cat(s);
		t->cat(data[i].toStr());
	}
	return t->data;
}

text array::toText(char *s = ", ") {
	int l = size();
	text *t = new text();
	for(int i = 0; i < l; i++) 	{
		if(i>0) t->cat(s);
		t->cat(data[i].toStr());
	}
	return t;
}


#ifdef DEBUG
void array::image()
{
 int l = (int) size();
 text x = "image[" + text(l) + "] (\n";
 char buf[64];

 for(int i = 0; i < l; i++)
 {
   x += key(i);
   x += ":";
   x += value(i).toText();
   x += " (";
   x += value(i).toText().length();
   x += ") ";
   x += text::isNumber(keyList[i].data);
   x += " adr=";
   sprintf(buf, "%ul", &data[i]);
   x += text(buf);
   x += "\n";
 }
 x += ")";

 std::cout << x << std::endl;
}
#endif

std::ostream& operator<< (std::ostream& out, array d)
{
  return out << d.toText().data;
}

void array::load(const char *name) {
    std::ifstream input(name);
    this->clear();
	for(std::string line; getline(input, line); ) {
		push(line.c_str());
	}
	input.close();
	return;
}

bool array::store(const char * name)
{
    std::ofstream f;
    f.open(name);
	for(int i = 0; i < size(); i++)	{
        f << data[i].toStr() << "\n";
	}
	f.close();
	return true;
}

array *array::to(const char *s) { return to(text(s)); }
array *array::to(text t)
{
   int k = findKey(t);
   if(k == -1)
   {
     k = size();
     keyList.push_back(text(k));
     data.push_back(dyn(new array()));
     freeKey = size();
   }

   return data[k].toArray();
}

array *array::to(int n)
{
  if(n >= freeKey)
  {
    freeKey = n + 1;
  }
  return to(text(n));
}

dyn &array::operator[](int n)
 {
   int k = findKey(n);
   if(k == -1)
   {
     k = size();
     keyList.push_back(text(n));
     data.push_back(dyn(false));
     if(n > freeKey) freeKey = n + 1;
   }
   return *(data.begin() + k);
 }

dyn & array::operator[](const char *s) { return operator[](text(s)); }
dyn & array::operator[](text t)
 {
   int k = findKey(t);
   if(k == -1) {
     k = size();
     keyList.push_back(t);
     data.push_back(dyn(false));
   }
   return *(data.begin() + k);
 }

dyn &array::at(int n)  {
   int k = findKey(n);
   if(k == -1)  {
     dyn::error = true;
     return nullelement;
   }
   return data[k];
 }
 
 dyn &array::at(const char *s) { return at(text(s)); }
 
 dyn &array::at(text t) {
   int k = findKey(t);
   if(k == -1)  {
     dyn::error = true;
     return nullelement;
   }
   return data[k];
 }
 
dyn &array::at(dyn d) { 
    if(d.isText())  return at(d.toText()); 
    return at(d.toInt());
} 

void array::setAt(int idx, dyn *a)
 {
   int k = findKey(idx);
   if(k != -1)
   {
     // the dyn held at k, in current array, replaced by a
     data[k].set(*a);
     return;
   }
   pushAdd(idx, *a);
 }

void array::setAt(int idx, dyn a)
 {
   int k = findKey(idx);
   if(k != -1)
   {
     data[k].set(a);
     return;
   }
   pushAdd(idx, a);
 }

void array::setAt(text t, dyn a)
 {
   int k = findKey(t);
   if(k != -1)
   {
     data[k].set(a);
     return;
   }
   pushAdd(t, a);
 }

void array::setAt(text t, dyn *a)
{
   setAt(t, *a);
}


void array::boundError(const text &t)
{
  std::cout << "Out of boundary " << t << std::endl;
}


//============= DICT

dict::dict()
{
  freeKey = 0;
  indexer = 0;
}


dict::dict(const dict &d) { set(d); }

dict::dict(dict *d) { set(d); }

void dict::set(const dict &a)  {
  keyList.assign(a.keyList.begin(), a.keyList.end());
  data.assign(a.data.begin(), a.data.end());
  freeKey = a.freeKey;
  indexer = 0;
}
void dict::set(dict *a) {
  keyList.assign(a->keyList.begin(), a->keyList.end());
  data.assign(a->data.begin(), a->data.end());
  freeKey = a->freeKey;
  indexer = 0;
}


void dict::erase(const char *s)
{
   for(unsigned i = 0; i < keyList.size(); i++)
   {
     //printf("%s = %s:%s\n", s, keyList[i].data, data[i].toStr());
     if(strcmp(s, keyList[i].data) == 0)
     {
       keyList.erase(keyList.begin() + i);
       data.erase(data.begin() + i);
       return;
     }
   }
}

dyn & dict::operator[](const text &t) { return operator[](t.data); }
dyn & dict::operator[](const char *s)
{
  int k = size();
  for(int i = 0; i < k; i++)
  {
     if(strcmp(s, keyList[i].data) == 0)
     {
       return data[i];
     }
  }
  keyList.push_back(text(s));
  data.push_back(dyn(false));
  return *(data.begin() + k);
}

void dict::setAt(const text &k, dyn a) { setAt(k, (dyn *) &a); }
void dict::setAt(const text &k, dyn *v)
 {
   int i = findKey(k.data);
   if(i == -1)
   {
     push(k,v);
     return;
   }
   data[i].set(v);
 }


void dict::insert(int n, dyn d)
{
  if(n < 0) n = size() + n;
  if(n < 0) return;
  
  bool TA = ((d.flag == TYPE_ARRAY) || (d.flag == TYPE_DICT));
  
  if(n >= size())
  {
    if(TA) add(d.a);
    else  push(d);
    return;
  }

  if(TA)
  {
   keyList.insert(keyList.begin() + n, d.a->keyList.begin(), d.a->keyList.end());
   data.insert(data.begin() + n, d.a->data.begin(), d.a->data.end());
  }
  else
  {
   keyList.insert(keyList.begin() + n, text(65000));
   data.insert(data.begin() + n, d);
  }
  pack();

}

void dict::unshift(text t, dyn a)
{
    erase(t.data);
    keyList.insert(keyList.begin(), t);
    data.insert(data.begin(), a);
    pack();
}

void dict::kSort(int x) { 
    std::vector<text> k = keyList;
    std::vector<dyn> v;
    std::sort(k.begin(), k.end());
    for(text key : k) {
        for(int i = 0; i < keyList.size(); i++)  {
            if(strcmp(key.data, keyList[i].data) == 0) {
                v.push_back(data[i]);
            }
        }
    }
    keyList = k;
    data = v;
}


