/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/



#ifndef _ARRAY_64
#define _ARRAY_64


#define TYPE_UNKNOW 0
#define TYPE_INT 1
#define TYPE_NUMBER 2
#define TYPE_NATURAL 3
#define TYPE_REAL 4
#define TYPE_BYTE  5
#define TYPE_BOOL 6
#define TYPE_LONG 7
#define TYPE_DOUBLE 8
#define TYPE_COMPLEX 9
#define TYPE_TEXT 16
#define TYPE_ARRAY 32
#define TYPE_DICT  33
#define TYPE_PAIR 34
#define TYPE_FILE 40
#define TYPE_DIR  41
#define TYPE_OBJECT 50
#define TYPE_TARRAY 60

extern int _default_capacity64;

/**
   Fast indexer for typed arrays of 32 bits objects:
   integers, strings of chars, pointers on objects...
  */

class _array64
{
 static const char *nothing;
 public:
 double *data;

/**
  Number of elements (less or equal to the capacity)
  */
 int len;

/**
   Default capacity at creation of the typed array
  */
 int capacity;

/**
  Used by the builtin iterator
  */
 int indexer;

 int type;

 /**
   default constructor
   */

 _array64();
 _array64(int);

 /**
   copy constructor
   */
 _array64(_array64 *);

 _array64(const _array64 &);

/**
  Constructor with type and one element
  */
 _array64(int, int);

/**
  Complete constructor with one element.
  In order, the type, the starting capcity, and the element
  */
 _array64(int, int, double);
 _array64(int, double);


/**
  Returns the number of elements
  */
 inline int size() { return len; }

/**
  Returns true if the array is empty
  */
 inline bool empty() { return len == 0; }


 void resize(int);

 /**
   Clears the content.
   */
 void clear();


 int compare(double, double);
 bool same(double, double);

 void set();
 void set(int);
 void set(int, int);
 void set(const _array64 &);
 void set(_array64 *);
 //void set(double);

 void operator=(const _array64 &);
 void operator=(_array64 *);
 void operator=(double);
 void operator=(natural);
 void operator=(int);
 
 void setAt(int, double);
 inline void setAt(int i, bool b) { erase(i); }  // set nil
 inline void setAt(int i, int n) { setAt(i, (double) n); }
 inline void setAt(int i, natural n) { setAt(i, (double) n); }
 inline void setAt(int i, const char *s) { setAt(i, text::toNumber(s)); }
 //void setAt(int, const char *);
 inline void setAt(int i, text &t) { setAt(i, t.toNumber()); }

 /**
   Adds a second array having the same type.
   */
 void add(_array64);

 double *operator[](int);
 double *operator[](const char *);
 double at(int);
 double at(text);
 double at(const char *);

 bool operator==(_array64);
 bool operator==(bool);

 inline bool operator!=(_array64 a) { return !operator==(a); }
 inline bool operator!=(bool b)     { return !operator==(b); }


 bool operator<(_array64);
 bool operator>(_array64);

 inline bool operator>=(_array64 a) { return !operator<(a); }
 inline bool operator<=(_array64 b) { return !operator>(b); }

 _array64 operator+(_array64);
 _array64 operator+(double);
 inline _array64 operator+(text x)    { return operator+(x.toNumber()); }
 inline _array64 operator+(int x)     { return operator+((double) x); }
 inline _array64 operator+(natural x) { return operator+((double) x); }
 inline _array64 operator+(char *x)   { return operator+(text::toNumber(x)); }


 _array64 operator-(_array64);
 _array64 operator-(double);
 inline _array64 operator-(text x)    { return operator-(x.toNumber()); }
 inline _array64 operator-(int x)     { return operator-((double) x); }
 inline _array64 operator-(natural x) { return operator-((double) x); }
 inline _array64 operator-(char *x)   { return operator-(text::toNumber(x)); }

 _array64 operator&(_array64);
 _array64 operator&(double);
 inline _array64 operator&(text x)    { return operator&(x.toNumber()); }
 inline _array64 operator&(int x)     { return operator&((double) x); }
 inline _array64 operator&(natural x) { return operator&((double) x); }
 inline _array64 operator&(char *x)   { return operator&(text::toNumber(x)); }

 _array64 operator|(_array64);
 _array64 operator|(double);
 inline _array64 operator|(text x)    { return operator|(x.toNumber()); }
 inline _array64 operator|(int x)     { return operator|((double) x); }
 inline _array64 operator|(natural x) { return operator|((double) x); }
 inline _array64 operator|(char *x)   { return operator|(text::toNumber(x)); }

 _array64 operator^(_array64);

 inline void operator+=(_array64 x) { add(x); }
 inline void operator+=(double x)    { push(x);}
 inline void operator+=(int x)       { operator+=((double) x); }
 inline void operator+=(natural x)   { operator+=((double) x); }
 inline void operator+=(text t)      { operator+=(t.toNumber()); }
 inline void operator+=(char *x)     { operator+=(text::toNumber(x)); }

 void operator-=(_array64);
 void operator-=(double);
 inline void operator-=(int x)       { operator-=((double) x); }
 inline void operator-=(natural x)   { operator-=((double) x); }
 inline void operator-=(text t)      { operator-=(t.toNumber()); }
 inline void operator-=(char *x)     { operator-=(text::toNumber(x)); }

 void operator&=(_array64);
 void operator&=(double);
 inline void operator&=(int i)       { operator&=((double) i); }
 inline void operator&=(natural i)   { operator&=((double) i); }
 inline void operator&=(text t)      { operator&=(t.toNumber()); }
 inline void operator&=(char *s)     { operator&=(text::toNumber(s)); }

 void operator|=(_array64);
 void operator|=(double);
 inline void operator|=(int i)       { operator|=((double) i); }
 inline void operator|=(natural i)   { operator|=((double) i); }
 inline void operator|=(text t)      { operator|=(t.toNumber()); }
 inline void operator|=(char *s)     { operator|=(text::toNumber(s)); }

 void operator^=(_array64);

 bool operator!();

 //===== util
 bool in(double);
 signed int hasStrKey(const text &);
 signed int findKey(const text &);
 signed int findKey(int);
 signed int findValue(double);

/**
  Finds the position of an element inside the array.
  Return the position or -1 if not found.
  */
 inline int find(double x) { return findValue(x); }

/**
  Displays the array
  */
 void display();
 void display(int);


/**
  Joins the elements of the array into a string.
  Numbers are converted to strings of digits.
  */
 char *toStr();
 char *toStr(int);
 
 
/**
  Joins the elements of the array into a text with separator.
  */
 char *toString(const char *s = ",");
 text toText(const char *s = ","); 


 //===== iterator

/**
  Points out the indexer at the first element.
  */
 double begin();

/**
  Points out the indexer at the last element.
  */
 double end();

/**
  Restarts the iterator at first element.
  */
 void reset();
 bool outOfBound();
 bool bound();
 bool inBound();

/**
  Returns the current element and increments the indexer to the next position.
  */
 double inc();

/**
  Returns the current element and decrements the indexer to the previous position.
  */
 double dec();

/**
  Returns the current position of the iterator. Use index() instead.
  */
 int key();
 int key(int);

/**
  Returns the currently pointed out element.
  */
 double value();
 double value(int);
 double current();

/**
  Return the current location of the indexer.
  */
 int index();

 //===== stack

/**
  Pushes a double at the end of the array.
  */
 void push(double n);


/**
  Pushes a natural at the end of the array.
  */
 inline void push(natural n) { push((double) n); }



 void pushUnic(double);

/**
  Returns the last element, and remove it.
  */
 double pop();

/**
  Returns the first element, and remove it.
  */
 double shift();

/**
  Inserts a double at beginning.
  */
 void unshift(double);

/**
  Inserts a natural at beginning.
  */
 inline void unshift(natural i) { unshift((double) i); }


 //===== ranges

/**
  Inserts an object at position.
  */
 void insert(int, double);

/**
  Inserts an array at position. The elements of the array are inserted,
  not the array itself.
 */
 void insert(int, _array64);


/**
  Removes a range of elements for the first to the last given positions,
  included. Use splicing instead.
  */
 void erase(int, int);

/**
  Removes the element at position.
  */

 void erase(int);
 _array64 slice(int, int, int);
 inline _array64 slice(int f) { return slice(f, len, 0); }

/**
  Extracts a sub-array from the array.
  Parameters are the position of the first and last element included.
  */
 inline _array64 slice(int f, int l) { return slice(f, l, 1); }

 void splice(int);

/**
  Erases a range of elements.
  Parameters are the position of the first and last element included.
  */
 void splice(int, int);

 void splice(int, int, int);
 void splice(int, int, _array64, int);
 void splice(int, int, double, int);
 void splice(int, int, bool, int);

 //===== transform

 /**
   Remove all null records from the array
   */
 void pack();

 void swapData(int, int);
 void dynSort(int,int);

/**
  Reverses the content of the array
  */
 _array64 reverse();

/**
  Sorts the array in ascending order.
  */
 void sort(int f = 0);
 void aSort(int f = 0);
 void kSort(int f = 0) { return; }

/**
  Removes doubloons
  */
 _array64 unique();

/**
  This method returns the array itself and is useless.
  */
 _array64 values();

/**
  Returns the indices as an array of integers.
  */
 _array64 keys();

 //===== operations

/**
  Copies into another array.
  */
 void clone(_array64 *);

/**
  Returns the smaller element.
  If elements are texts, they are considered to represent integers,
  and non-digits text have 0 value.
  */
 double min();

/**
  Returns the higher element.
  If elements are texts, they are considered to represent integers.
  */
 double max();

/**
  Returns the sum of elements.
  If elements are texts, they are considered to represent integers.
  */
 double sum();

/**
  Joins the elements into a text when they are string and thus, for an
  array of texts only.
  */
 text join(const char *);

 //===== I/O

/**
  Fills an array from a file (one element per line).
  */
 void load(const char *);

/**
 Store an array into file. The end of line code must be added
 to elements of the array. Elements are stored as text if they are texts,
 as binary if they are scalars. Object are not handled for now.
 */
 bool store(const char *);

 static void boundError(const text &);
};

std::ostream& operator<< (std::ostream &, _array64 &);
//_array64* operator+(_array64*, double);

#endif

