/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#include <stdlib.h>
#include <stdio.h>
#include <float.h>
#include <stdarg.h>
#include <cstring>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <limits.h>

#include "text.hpp"
#include "intarr.hpp"

/** \file dynarr.cpp
  File of the typed array 32 bits library
  */

/** \class _intArr
   Base class for typed array of 32 bits objects
   including integers, string of chars, pointers on objects...
 */


int _default_capacityInt = 16;

//=====  Constructors

/**
   Basic constructor
  */
  
_intArr::_intArr() { 
    set(); 
}

_intArr::_intArr(int n) { 
    set(); 
    push(n);
}

_intArr::_intArr(int t, int n) { 
    set(t); 
    push(n);
}

_intArr::_intArr(_intArr *a)  { set(a); }

_intArr::_intArr(const _intArr &a) { set(a); }



void _intArr::set() { 
    set(_default_capacityInt); 
}

void _intArr::set(int si) {
    data = (int *)  new char[si * sizeof(int)];
    capacity = si;
    len = 0;
    indexer = 0;
}

void _intArr::set(const _intArr &model) {
    set(model.capacity);
    for(int i  = 0; i < model.len; i++) {
      push(model.data[i]);
    }
}

void _intArr::set(_intArr *model) {
    set(model->capacity);
    for(int i  = 0; i < model->len; i++)  {
      push(model->data[i]);
    }
}


//===== Utilities

void _intArr::clear() {
  set(_default_capacityInt);
}

void _intArr::resize(int i) {
  if(i >= len)  {
    if(i >= capacity)  {
       capacity = i + 16;
       data = (int *) realloc(data, capacity * sizeof(int));
    }
    len = i + 1;
  }
}

char *_intArr::toStr() {
    text t = join();
    return t.data;
}

char *_intArr::toStr(int n) {
  char *buffer = (char *) new char[16];
  sprintf(buffer, "%d", n);
  return buffer;
}

void _intArr::pack() {
    int i, j;

    j = 1;
    for(i = 0; i < len; i++) {
        if(data[i] == 0) {
            while(data[j] == 0) {
                j++;
            }    
            if(j >= len) {
                len = i;
                return;
            }
            data[i] = data[j];
        }
        j++;
    }
    len = i;
    return;
}




//===== GET

int & _intArr::operator[](int i)  {
   if(i >= len) {
     resize(i);
     data[i] = 0;
   }
   return data[i];
 }


int _intArr::at(int n)  {
   if(n < len) return data[n];
   return -1;
}


void _intArr::setAt(int i, int n) {
   if(i < len) data[i] = n;
   else push(n);
 }


void _intArr::add(_intArr x) {
  int l = x.size();
  for(int i = 0; i < l; i++) {
     push(x.data[i]);
  }
}


void _intArr::clone(_intArr *y)  {
   y->set(*this);
 }


//===== OPERATORS

void _intArr::operator=(const _intArr &a)  { 
    set(a); 
}

void _intArr::operator=(_intArr *a) { 
    set(a);
}

void _intArr::operator=(int i) {
  set();
  push(i);
}

_intArr _intArr::operator+(_intArr x) {
   _intArr y = _intArr(this);
   y.add(x);
   return y;
 }

 _intArr _intArr::operator+(int x) {
   _intArr y = _intArr(this);
   y.push(x);
   return y;
 }

_intArr _intArr::operator-(_intArr b) {
   int i,j;
   int l = b.size();
   _intArr y;
   y.set(len);
   int x;

   for(i = 0; i < len; i++)  {
     x = data[i];
     for(j = 0; j  < l; j++) {
       if(x == b.data[j]) goto skip;
     }
     y.push(x);
skip:;     
   }
   return y;
};

_intArr _intArr::operator-(int x) {
   _intArr y;
   y.set(len);
   for(int i = 0; i < len; i++)  {
     if(data[i] != x)  y.push(data[i]);
   }
   return y;
}

 _intArr _intArr::operator&(_intArr b)  {
   int i,j;
   int l = b.size();
   _intArr y;
   int m = (l < size()) ? l : size();
   y.set(l);
   int x;

   for(i = 0; i < size(); i++)  {
     x = data[i];
     for(j = 0; j  < l; j++)  {
       if(x == b.data[j])  y.push(x);
     }
   }
   return y;
 };

_intArr _intArr::operator&(int x) {
  _intArr y = _intArr(this);
  if(findValue(x) > -1) y.push(x);
  return y;
}

_intArr  _intArr::operator|(_intArr b) {
  int i, l = b.size();
  _intArr y;
  y.set(len + l);

  for(i = 0; i < len; i++) y.pushUnic(data[i]);
  for(i = 0; i < l; i++)   y.pushUnic(b.data[i]);

  return y;
}

_intArr _intArr::operator|(int x)
 {
   _intArr y = _intArr(this);
   y.pushUnic(x);
   return y;
 }

_intArr _intArr::operator^(_intArr b) {
  int i;
  _intArr y;
  y.set(len);

  for(i = 0; i < len; i++) {
    if(b.findValue(data[i]) == -1)
      y.pushUnic(data[i]);
  }
  for(i = 0; i < b.size(); i++)  {
    if(findValue(b.data[i]) == -1)
      y.pushUnic(b.data[i]);
  }
  return y;
}

void _intArr::operator+=(_intArr x) { add(x); }
void _intArr::operator+=(int x)   { push(x);}


 void _intArr::operator-=(_intArr x) {
   _intArr y = _intArr(this);
   set(y.operator-(x));
 }

 void _intArr::operator-=(int x) {
   _intArr y = _intArr(this);
   set(y.operator-(x));
 }

void _intArr::operator&=(_intArr b) {
   _intArr y = operator&(b);
   set(&y);
 }

void _intArr::operator&=(int x) {
   int flag = findValue(x) > -1;
   clear();
   if(flag) push(x);
 }


void _intArr::operator|=(_intArr b) {
   _intArr  y = operator|(b);
   set(&y);
 }

void _intArr::operator|=(int x) {
   pushUnic(x);
}

void _intArr::operator^=(_intArr b)  {
   _intArr  y = operator^(b);
   set(&y);
 }


bool _intArr::operator==(_intArr b)  {
   if(b.size() != len) return false;
   for(int i = 0; i < len; i++)  {
     if(data[i] != b.data[i]) return false;
   }
   return true;
 }

 bool  _intArr::operator==(bool x)  { return (size() != 0) == x; }

 bool _intArr::operator<(_intArr a)  {
   int l = a.size();
   if(len < l) return true;
   if(len > l) return false;
   for(int i = 0; i < l; i++)  {
    if(data[i] >= a.data[i]) return false;
   }
   return true;
 }

 bool _intArr::operator>(_intArr a)  {
   int l = a.size();
   if(len < l) return false;
   if(len > l) return true;
   for(int i = 0; i < l; i++)  {
     if(data[i] <= a.data[i]) return false;
   }
   return true;
 }

bool _intArr::operator!() { 
    return len == 0; 
}



//===== util

 bool _intArr::in(int x)  {
  for(int i = 0; i < size(); i++)  {
   if(data[i] == x) return true;
  }
  return false;
 }

signed int _intArr::findKey(int n) { return n; }

signed int _intArr::findValue(int x) {
  for(int i = 0; i < size(); i++)  {
   if(data[i] == x) return i;
  }
  return -1;
}

void _intArr::display() { 
    display(0); 
}

void _intArr::display(int tab) {
 std::cout << "Array of int(\n";

 for(int i = 0; i < len; i++) {
   for(int j = 0; j < tab; j++) std::cout << "     ";
   std::cout << data[i] << std::endl;
 }
 for(int j = 0; j < tab; j++) std::cout << "     ";
 std::cout << ")\n";
}

 //===== iterator

 void _intArr::reset() { indexer = 0; }
 bool _intArr::outOfBound() { return indexer >= size(); }
 bool _intArr::bound()      { return(++indexer < size()); }
 
 bool _intArr::inBound() {
  if(indexer < size()) { 
      return true;
  }
  return false;
 }

 int  _intArr::inc()  {
  if(++indexer < size()) return data[indexer];
  indexer = size();
  return -1;
 }

 int  _intArr::dec() {
  if(--indexer >= 0) return data[indexer];
  indexer = 0;
  return -1;
 }

 int _intArr::current() { 
     return value(indexer); 
 }
 
 int _intArr::value() { 
     return value(indexer); 
 }
 
 int _intArr::value(int i)  {
   if(i >=0 && i < size())  return data[i];
   return -1;
 }

 int _intArr::key() {
   if(indexer >= 0 && indexer < size())  return indexer;
   return -1;
  }

 int _intArr::key(int i)
 {
   if(i >= 0 && i < size())  return i;
   return -1;
 }

 int _intArr::index()  {
   if(indexer >=0 && indexer < size()) return indexer;
   return -1;
 }

int _intArr::begin()  {
   indexer = 0;
   if(size() > 0) return data[0];
   return -1;
}

int _intArr::end() {
   indexer = 1;
   if(size() > 0)    {
     indexer = size() -1;
     return data[indexer];
   }
   return -1;
}

//===== stack


 void _intArr::push(int x) {
    resize(len);
    data[len-1] = x;
 }


 void _intArr::pushUnic(int x) {
     if(findValue(x) == -1) {
       push(x);
     }
 }

 int _intArr::pop()  {
   indexer = 0;
   if(!empty()) {
     len --;
     return data[len];
   }
   return -1;
 }

int _intArr::shift() {
  if(empty())  return -1;
  else  {
   int x = data[0];
   erase(0);
   return x;
  }
}

void _intArr::unshift(int x) {
  insert(0, x);
}

void _intArr::insert(int n, int x) {
  if(n < 0) n += len;
  if(n < 0) return;
  if(n >= len) {
    push(x);
    return;
  }

  int y = data[len - 1];
  for(int i = len - 1; i > n; i--) {
    data[i] = data[i - 1];
  }
  data[n] = x;
  push(y);
}

void _intArr::insert(int n, _intArr d) {
  if(n < 0) n += len;
  if(n < 0) return;
  if(n >= len)  {
    add(d);
    return;
  }

  int si = len + d.size();

  _intArr x;
  x.set(si);
  if(n > 0) x.add(slice(0, n, 0));
  x.add(d);
  x.add(slice(n));
  set(x);
}

//===== RANGE


 // erase from first to last included
 void _intArr::erase(int first, int last)  {
   int k, l;
   if(first > len) return;
   if(last >= len) {
      len = first + 1;
      return;
   }

   l = len;
   
   int nl = len - (last - first + 1);
   if(nl < 0) nl = 0;
   k = last;

   while(first < len) {
     k++;
     if(k < l) {
       data[first] = data[k];
     }
     else {
       data[first] = -1;
     }
     first++;
   }
   len = nl;
 }

 void _intArr::erase(int pos)  {
    if(pos >= len) return;
    len--;
    while(pos < len)  {
       data[pos] = data[pos + 1];
       pos++;
    }
    data[len] = -1;
    return;
 }

 _intArr _intArr::slice(int f, int l, int incl)  {
   _intArr x;

   l += incl;
   l--;
   if(l < 0) l += len;
   if(l < 0) goto bye;
   if(l >= len) l = len - 1;

   if(f < 0) f += len;
   if(f < 0) f = 0;
   if(l < f) goto bye;
   if(f >= len) goto bye;

   x.set(l - f + 1);

   if(empty()) goto bye;

   while(f <= l)  {
     x.push(data[f++]);
   }
bye:
   return x;
 }

 void _intArr::splice(int b, int e)  {
   if(b < 0) b += len;
   if(b < 0) b = 0;
   if(e < 0) e += len;
   if(e < 0) return;
   if(e >= len) e = len - 1;
   if(e < b) return;
   erase(b, e);
 }

 void _intArr::splice(int b, int e, int inc)  {
   splice(b, e + inc - 1);
 }

 void _intArr::splice(int f)  { erase(f); }

// Content of array a is inserted or replace the content of the interval defined

 void _intArr::splice(int i, int last, _intArr a, int incl) {
  if(empty())   { operator=(a); return; }
  if(a.empty()) { erase(i, last + incl - 1); return; }
  if(last < 0) last += len;
  if(i < 0) i += len;
  if(i < 0) i = 0;
  last += incl;
  last--;
  if(last >= len) last = len - 1;
  if(i >= len)  {
    add(a);
    return;
  }
  if(last < i) return;
  if(last < 0) return;
  if((last + 1) == i) {
    insert(i, a);       // providing not a.empty() as tested above
    return;
  }
  if(a.empty())  {
    erase(i, last);
    return;
  }

  int j = 0;
  while((i <= last) && (j < a.size()))  {
    data[i] = a.data[j];
    i++;
    j++;
  }
  if(last >= i) erase(i, last);
  if(j < a.size())  {
    insert(i, a.slice(j));
  }
 }


 void _intArr::splice(int i, int last, int x, int incl)
 {
  if(i < 0) i += len;
  if(i < 0) i = 0;
  if(last < 0) last += len;
  last += incl;
  last --;
  if(last >= len) last = len - 1;
  if(last < i) return;
  if(last < 0) return;

  if(i >= len)  { push(x); return; }

  data[i] = x;
  if(last > i)  {
    erase(i + 1, last);
  }
 }

//===== TRANSFORM

/*
  if data is a not a number, search if exists further
  if yes, wait for the next occurence to push
  if not a number push it providing not already pushed
*/

_intArr _intArr::unique() {
  _intArr y;
  y.set(len);

  for(int i = 0; i < len; i++)  {
    y.pushUnic(data[i]);
  }

  return y;
}


void _intArr::swapData(int a, int b) {
  int x;
  x = data[a];
  data[a] = data[b];
  data[b] = x;
}


void _intArr::dynSort(int first,int last) {
 int f=first;
 int l=last;
 int item = data[(f + l) / 2];

 do  {
   while(data[f] < item) f++;
   while(item < data[l]) l--;
   if(f <= l)  {
     swapData(f, l);
     f++;
     l--;
   }
 } while(f <= l);

 if(first < l)  dynSort(first, l);
 if(f < last)   dynSort(f, last);
 return;
}

void _intArr::sort(int i)  { dynSort(0, len - 1); }
void _intArr::aSort(int i) { dynSort(0, len - 1); }


_intArr  _intArr::keys()
{
  _intArr x;
  x.set(len);
  for(int i = 0; i < len; i++)  {
    x.push(i);
  }
  return x;
}

_intArr _intArr::values() {
  return *this;
}

//===== UTILITIES

int _intArr::min() {
  int m = INT_MAX;
  for(int i = 0; i < len; i++)  {
    if(data[i] < m) m = data[i];
  }
  return m;
}

int _intArr::max() {
  int m = INT_MIN;
  for(int i = 0; i < len; i++) {
    if(data[i] > m) m = data[i];
  }
  return m;
}


int _intArr::sum() {
  int m = 0;
  for(int i = 0; i < len; i++) {
    m += data[i];
  }
  return m;
}


_intArr _intArr::reverse()
{
  _intArr y;
  y.set(len);
  for(int i = len - 1; i >= 0; i--) {
    y.push(data[i]);
  } 
  return y;
}

char *_intArr::toString(const char *s) {
	text *t = new text();
    t->cat("(");
	for(int i = 0; i < len; i++) 	{
		if(i > 0) t->cat(s);
		t->cat(toStr(data[i]));
	}
	t->cat(")");
	return t->data;
}

text _intArr::toText(const char *s) {
    return toString(s);
}

text _intArr::join(const char *s) {
  return text(toString(s));
}


std::ostream& operator<< (std::ostream& out, _intArr &d) {
  return out << d.toStr();
}

void _intArr::load(const char *name) {
    std::ifstream input(name);
    this->clear();
	for(std::string line; getline(input, line); ) {
		push(atoi(line.c_str()));
	}
	input.close(); 
	return;
}

bool _intArr::store(const char * name) {
    std::ofstream f;
    f.open(name);
	for(int i = 0; i < size(); i++)	{
	    std::stringstream out;
        out << data[i];
        f <<  out.str() << "\n";
	}
	f.close();
    return true;
}


void _intArr::boundError(const text &t) {
  std::cout << "Out of boundary " << t.data << std::endl;
}

