/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#ifndef _INT_ARRAY
#define _INT_ARRAY


extern int _default_capacityInt;

/**
   Fast indexer for typed arrays of integers
  */

class _intArr 
{
 static const char *nothing;
 
 public:
 int *data;

/**
  Number of elements (less or equal to the capacity)
  */
 int len;

/**
   Default capacity at creation of the typed array
  */
 int capacity;

/**
  Used by the builtin iterator
  */
 int indexer;

 /**
   default constructor
   */

 _intArr();
 _intArr(int);          // one number 
 _intArr(int, int);     // capacity and one number
 
 /**
   copy constructors
   */
 _intArr(_intArr *);
 _intArr(const _intArr &);

/**
  Returns the number of elements
  */
 inline int size() { return len; }

/**
  Returns true if the array is empty
  */
 inline bool empty() { return(len == 0); }


 void resize(int);

 /**
   Clears the content.
   */
 void clear();



 void set();
 void set(int);
 void set(const _intArr &);
 void set(_intArr *);

 void operator=(const _intArr &);
 void operator=(_intArr *);
 void operator=(int);

 void setAt(int, int);
 inline void setAt(int i, bool b) { erase(i); }  // set nil

 /**
   Adds a second array having the same type.
   */
 void add(_intArr);

 int &operator[](int);
 int at(int);

 bool operator==(_intArr);
 bool operator==(bool);
 inline bool operator!=(_intArr a) { return !operator==(a); }
 inline bool operator!=(bool b)   { return !operator==(b); }
 bool operator<(_intArr);
 bool operator>(_intArr);
 inline bool operator>=(_intArr a) { return !operator<(a); }
 inline bool operator<=(_intArr b) { return !operator>(b); }

 _intArr operator+(_intArr);
 _intArr operator+(int x);


 _intArr operator-(_intArr);
 _intArr operator-(int x);

 _intArr operator&(_intArr);
 _intArr operator&(int x);

 _intArr operator|(_intArr);
 _intArr operator|(int x);

 _intArr operator^(_intArr);

 void operator+=(_intArr);
 void operator+=(int x);

 void operator-=(_intArr);
 void operator-=(int x);

 void operator&=(_intArr);
 void operator&=(int i);

 void operator|=(_intArr);
 void operator|=(int i);

 void operator^=(_intArr);

 bool operator!();

 //===== util
 signed int findKey(int);
 signed int findValue(int);
 
 bool in(int);

/**
  Finds the position of an element inside the array.
  Return the position or -1 if not found.
  */
 inline int find(int x) { return findValue(x); };

/**
  Displays the array
  */
 void display();
 void display(int);


/**
  Joins the elements of the array into a string.
  Numbers are converted to strings of digits.
  */
 char *toStr();

 char *toStr(int);

/**
  Joins the elements of the array into a text.
  */
 text toText();

 //===== iterator

/**
  Points out the indexer at the first element.
  */
 int begin();

/**
  Points out the indexer at the last element.
  */
 int end();

/**
  Restarts the iterator at first element.
  */
 void reset();
 bool outOfBound();
 bool bound();
 bool inBound();

/**
  Returns the current element and int rements the indexer to the next position.
  */
 int inc();

/**
  Returns the current element and decrements the indexer to the previous position.
  */
 int dec();

/**
  Returns the current position of the iterator. Use index() instead.
  */
 int key();
 int key(int);

/**
  Returns the currently pointer out element.
  */
 int value();
 int value(int);
 int current();

/**
  Return the current location of the indexer.
  */
 int index();

 //===== stack


/**
  Pushes an integer at the end of the array.
  */
 void push(int);

 void pushUnic(int);

/**
  Returns the last element, and remove it.
  */
 int pop();

/**
  Returns the first element, and remove it.
  */
 int shift();

/**
  Inserts an integer at beginning.
  */
 void unshift(int);

 //===== ranges

/**
  Inserts an object at position.
  */
 void insert(int, int);


/**
  Inserts an array at position. The elements of the array are inserted,
  not the array itself.
 */
 void insert(int, _intArr);


/**
  Removes a range of elements for the first to the last given positions,
  included. Use splicing instead.
  */
 void erase(int, int);

/**
  Removes the element at position.
  */

 void erase(int);
 
/**
    Extract a sub array
    */
 
 _intArr slice(int, int, int);
 inline _intArr slice(int f) { return slice(f, len, 0); }

/**
  Extracts a sub-array from the array.
  Parameters are the position of the first and last element included.
  */
 inline _intArr slice(int f, int l) { return slice(f, l, 1); }

 void splice(int);

/**
  Erases a range of elements.
  Parameters are the position of the first and last element included.
  */
 void splice(int, int);

 void splice(int, int, int);
 void splice(int, int, int, int);
 void splice(int, int, _intArr, int);
 void splice(int, int, bool, int);

 //===== transform

 /**
   Remove all null records from the array
   */
 void pack();

 void swapData(int, int);
 void dynSort(int,int);

/**
  Reverses the content of the array
  */
 _intArr reverse();

/**
  Sorts the array in ascending order.
  */
 void sort(int i = 0);
 void aSort(int i = 0);
 void kSort(int i = 0) { return; }

/**
  Removes doubloons
  */
  _intArr unique();

/**
  This method returns the array itself and is useless.
  */
 _intArr values();

/**
  Returns the indices as an array of integers.
  */
 _intArr keys();

 //===== operations

/**
  Copies into another array.
  */
 void clone(_intArr *);

/**
  Returns the smaller element.
  If elements are texts, they are considered to represent integers,
  and non-digits text have 0 value.
  */
 int min();

/**
  Returns the higher element.
  If elements are texts, they are considered to represent integers.
  */
 int max();

/**
  Returns the sum of elements.
  If elements are texts, they are considered to represent integers.
  */
 int sum();

/**
  Joins the elements into a text when they are string and thus, for an
  array of texts only.
  */
 text join(const char * = ",");

 //===== I/O

/**
  Fills an array from a file (one element per line).
  */
 void load(const char *);

/**
 Store an array into file. The end of line code must be added
 to elements of the array. Elements are stored as text if they are texts,
 as binary if they are scalars. Object are not handled for now.
 */
 bool store(const char *);

 char *toString(const char *s = ",");
 text toText(const char *s = ",");
 
 static void boundError(const text &);
};

std::ostream& operator<< (std::ostream &, _intArr &);
//_intArr* operator+(_intArr*, int );

#endif

