/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


/**
   Xml Checker - scriptol.com
   C++ wrapper to expat
   Build an automaton from callbacks that call it with a code
   according to events
   */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef UNIX
#include <unistd.h>
#else
#include <io.h>
#endif
#include <ctype.h>
#include <assert.h>
#include <math.h>
#include <string>
#include <vector>
#include <iostream>
#include <algorithm>

#ifndef GC_DISABLED
#include "gc_cpp.h"
#endif

#ifndef UNIX
#define strcasecmp stricmp
#endif


#include "text.hpp"
#include "dom.hpp"
#include "expat.h"
#include "sax.hpp"

int sax::depth = 0;
int sax::depthold = 0;
const char *sax::elementName;
const char *sax::commentText;
const char **sax::attributes;
const char *sax::dataText;
const char *sax::attributeName = NULL;
char *sax::attributeValue;
char *sax::searchName = NULL;
char *sax::foundValue;
int sax::len = 0;
bool sax::found = false;
void (*sax::automaton)(int);


void XMLCALL sax::startElement(void *data, const char *name, const char **attr)
 {
	elementName = name;
	attributes = attr;
	automaton(SAX_START);
	depth++;
 }

void XMLCALL sax::endElement(void *data, const char *name)
 {
	depth--;
	elementName = name;
	automaton(SAX_END);
 }

 // Test if a real content, this tools handle any char outside elements!

bool sax::valid(const char *str, int len)
 {
	int i;
	char c;

	for(i = 0; i < len; i++)
	{
		c = str[i];
		if(c == 13) continue;
		if(c == 10) continue;
		if(c == '\t') continue;
		if(c == ' ') continue;
		return true;
	}
	return false;
}


 // Processing the content

void XMLCALL sax::content(void *data, const char *thetext, int l)
 {
	if(! valid(thetext, l)) return;

	dataText = thetext;
	len = l;
	automaton(SAX_DATA);
 }

 // Handling a comment

void XMLCALL sax::comment(void *data, const char *cmt)
 {
	commentText = cmt;
	automaton(SAX_COMMENT);
 }

 
 
/* 
  Parse the whole document and throw events
*/ 

bool XMLCALL sax::parse(void (*fun) (int), const char *fname)
{
	automaton = fun;
	return parse(fname);
}

bool XMLCALL sax::parse(const char *fname)
 {
	FILE *fp;
	#define BUFFSIZE 8192
	char Buff[BUFFSIZE];

	XML_Parser p = XML_ParserCreate(NULL);
	if (! p)
	{
		fprintf(stderr, "Not enough memory for parser\n");
		exit(-1);
	}

	XML_SetElementHandler(p, startElement, endElement);
	XML_SetCharacterDataHandler(p, content);
	XML_SetCommentHandler(p, comment);

	fp = fopen(fname, "r");
	if(fp == NULL)
	{
		puts("File not found");
		exit(0);
	}

	for (;;)
	{
		int done;
		int len;

		len = fread(Buff, 1, BUFFSIZE, fp);
		if (ferror(fp))
		{
			fprintf(stderr, "Read error\n");
			return false;
		}
		done = feof(fp);

		if (XML_Parse(p, Buff, len, done) == XML_STATUS_ERROR)
		{
			fprintf(stderr, "Parse error at line %d:\n%s\n",
			XML_GetCurrentLineNumber(p),
			XML_ErrorString(XML_GetErrorCode(p)));
			return false;
		}

		if (done) break;
	}
	return true;
 }


// Display

void sax::display(int code)
 {
	int i;

	switch(code)
	{
	case SAX_START:
		for(i = 0; i < depth; i++)   printf("  ");
		printf("<%s", elementName);
		for (i = 0; attributes[i]; i += 2)
		{
			printf(" %s='%s'", attributes[i], attributes[i + 1]);
		}
		printf(" >\n");
		break;
	case SAX_END:
		for (i = 0; i < depth; i++)   printf("  ");
		printf("</%s>\n", elementName);
		break;
	case SAX_DATA:
		//for (i = 1; i < depth; i++)  printf("  ");
		for(i = 0; i < len; i++)
		{
			putchar(dataText[i]);
		}
		printf("\n");
		break;
	case SAX_COMMENT:
		for (i = 1; i < depth; i++)  printf("  ");
		printf("%s\n", commentText);
		break;
	default:
		break;
   }
   return;
 }


// Find data or attribute value
// - a searchName is set a string, searches for element
// - if attributeName is set NULL, returns data
// - else searches for for attributeName, return value

void sax::find(int code)
 {
	int i, l;
	const char *value;

	switch(code)
	{
	case SAX_START:
		if(found) break;   // first only
		if(attributeName == NULL) break;
		if(searchName != NULL)
		{
			if(strcasecmp(searchName, elementName) != 0) break;
		}

		for (i = 0; attributes[i]; i += 2)
		{
			if(strcasecmp(attributeName, attributes[i]) == 0)
			{
				value = attributes[i + 1];
				l = strlen(value);
				foundValue = (char *) GC_malloc(l + 1);
				strncpy(foundValue, value, l);
				foundValue[l] = '\0';
				found = true;
				break;
			}
		}
		break;
	case SAX_DATA:
		if(found) break;
		if(searchName == NULL) break;
		if(attributeName != NULL) break;
		printf("%s = %s\n", searchName, elementName);
		if(strcasecmp(searchName, elementName) == 0)
		{
			foundValue = (char *) dataText;
			found = true;
		}
		break;
	default:
		break;
	}
	return;
}


/*
 Build dom tree using sax-expat
*/

xelement *lxelement = new xelement();
int _sax_build_len = 0;
int _sax_build_size = 256;
char *_sax_buffer;

char *_sax_realloc(char *temp)
{
	int i;
	char *temp2 = (char *) GC_malloc_atomic(_sax_build_size);
	for(i = 0; i  <= _sax_build_len; i++)
	{
		temp2[i] = temp[i];
	}
	return temp2;
}

void _sax_build(int code)
{
	int i = 0;
	char *temp2;
	int l = 0;

	if(lxdom == NULL)
	{
		lxdom = new dom();
	}
	
	switch(code)
	{
	case SAX_START:
		lxelement = new xelement((char *) sax::elementName);
		_sax_build_len = 0;
		//printf("parent index %d\n", lxdom->index);

		if((sax::depth > sax::depthold) || (lxdom->getIndex() == 0))
		{
			lxdom->addChild(lxelement);
			//printf("child index %d\n", lxdom->index);
		}
		else
		{
			lxdom->addNext(lxelement);
		}
		for (i = 0; sax::attributes[i]; i += 2)
		{
			lxelement->addAttribute(sax::attributes[i], sax::attributes[i + 1]);
		}
		break;
	case SAX_END:
		if(_sax_build_len)
		{	
			lxelement->setData(_sax_buffer);
		}	
		if(sax::depth < sax::depthold)
		{
			lxdom->up();
		}
		sax::depthold = sax::depth;
		break;
	case SAX_DATA:
		l = sax::len;
		temp2 = (char *) sax::dataText;
		if(_sax_build_len == 0)
		{
			//printf("DATA: %d\n", l);
			_sax_build_size = 128;
			if((l + 1) > _sax_build_size) _sax_build_size = l + 1;
			_sax_buffer = (char *) GC_malloc_atomic(_sax_build_size);
			for(i = 0; i < l; i++)
			{
				_sax_buffer[i] = temp2[i];
			}
			_sax_build_len = l;
		}
		else
		{
			//printf("MORE: %d\n", l);
			if(l + _sax_build_len > _sax_build_size)
			{
				_sax_build_size += 128;
				_sax_buffer = _sax_realloc(_sax_buffer);
			}
			_sax_buffer[_sax_build_len] = '\n';
			_sax_build_len++;
			for(i = 0; i < l; i++)
			{
				_sax_buffer[i + _sax_build_len] = temp2[i];
			}
			_sax_build_len += l;
			
		}

		_sax_buffer[_sax_build_len] = '\0';
		//printf("%s\n", _sax_buffer);
		break;
	case SAX_COMMENT:
		if(sax::commentText != NULL)
		{
			text t = text("<!--") + text((char *) sax::commentText) + "-->";
			lxdom->addComment(t.data);
		}
		break;
	default:
		break;
	}
	return;
}


/**
  Function to parse an xml document into a Scriptol dom tree
  lxdom must be assigned an instance of dom
*/

bool SaxImport(const char *fname)
{
	return sax::parse(_sax_build, fname);
}


#define _SAX_STANDALONE
#ifdef SAX_STANDALONE

#include <dir.h>

bool LOG = false;
FILE *logfile;

void usage()
{
	puts("Xml Check - scriptet.com");
	puts("Syntax: xcheck [-view | -log] file.xml");
	exit(0);
}

void display(int code)
{
	sax::display(code);
}

char *changeExt(const char *fname, const char *newext)
{
	char *newname = new char[256];
	int len = strlen(fname);
	int i = 0;
	
	strcpy(newname, fname);
	
	for(i = len - 1; i--; i >= 0)
	{
		if(fname[i] == '.') break;
	}
	
	if(i >= 0) newname[i] = '\0';
	strcat(newname, newext);	

	return newname;
}

void createLog(const char *path)
{
	int pl = strlen(path);
	char buffer[255];

	// change extension

	char *logname = changeExt(path, ".log");
	
	logfile = fopen(logname, "w");
}

void message(const char *fname, const char *str)
{
	if(LOG)
	{
		fputs(fname, logfile);
		fputs(str, logfile);
		fputs("\n", logfile);
		return;
	}
	
	printf(fname);
	printf(str);
}


int main(int argc, char *argv[])
{
	const char *fname;
	bool VIEW = false;
	int offset = 1;
	
	if(argc < 2) usage();

	fname = argv[offset];
	if (strcmp(fname , "-view" ) == 0)
	{
		if(argc < 3)
		{
			usage();
		}
		offset+=1;
		fname = argv[offset];
		VIEW = true;
	}

	if (strcmp(fname , "-log" ) == 0)
	{
		if(argc < 3)
		{
			usage();
		}
		offset += 1;
		fname = argv[offset];
		LOG = true;
		createLog(argv[0]);
	}

	
	dom *mydom = new dom();
	lxdom = mydom;
	//sax::parse(display, fname);
	bool ok = sax::parse(_sax_build, fname);
	
	if(ok)
		message(fname, " is well-formed, ok...");
	else
		message(fname, " has errors...");

	if(VIEW) mydom->display();
	if(LOG) fclose(logfile);
	
	return 0;
}

#endif


#define _SAX_FINDER
#ifdef SAX_FINDER

int main(int argc, char *argv[])
{
  const char *fname;

  if(argc < 3)
  {
    puts("XCheck - scriptet.com");
    puts("Syntax: xcheck file.xml attribute-name");
    exit(0);
  }

  fname = argv[1];

  //sax::searchName = argv[2];
  sax::found = false;
  sax::searchName = NULL;
  sax::attributeName = argv[2];
  //sax::attributeName = NULL;
  sax::attributeName = "value";
  sax::parse(sax::find, fname);
  printf("Searching %s...\n", sax::searchName);
  printf("Content of %s is %s \n", sax::elementName, sax::dataText);
  //printf("Value of %s is %s \n", sax::attributeName, sax::foundValue);

  return 0;
}

#endif

