/*
The Scriptol library
Home page of the compiler: http://www.scriptol.org

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is the Scriptol library for the Scriptol compiler.
The Initial Developer of the Original Code is Denis Sureau at Scriptol.org.
*/


#ifndef _DYN_ARRAY
#define _DYN_ARRAY


#define TYPE_UNKNOW 0
#define TYPE_INT 1
#define TYPE_NUMBER 2
#define TYPE_NATURAL 3
#define TYPE_REAL 4
#define TYPE_BYTE  5
#define TYPE_BOOL 6
#define TYPE_LONG 7
#define TYPE_DOUBLE 8
#define TYPE_COMPLEX 9
#define TYPE_TEXT 16
#define TYPE_ARRAY 32
#define TYPE_DICT  33
#define TYPE_PAIR 34
#define TYPE_FILE 40
#define TYPE_DIR  41
#define TYPE_OBJECT 50
#define TYPE_ARRAY32 60
#define TYPE_ARRAY64 61
#define TYPE_INTARR 62

extern int _default_capacity32;

/**
   Fast indexer for typed arrays of pointers:
   strings of chars, pointers on objects...
  */



class _arr32 
{
 static const char *nothing;
 
 public:
 char **data;

/**
  Number of elements (less or equal to the capacity)
  */
 int len;

/**
   Default capacity at creation of the typed array
  */
 int capacity;

/**
  Used by the builtin iterator
  */
 int indexer;

 int type;

 /**
   default constructor
   */

 _arr32();
 _arr32(int);
 
 /**
   copy constructor
   */
 _arr32(_arr32 *);

 _arr32(const _arr32 &);

/**
 Constructor from an array of texts, with text type, size of the array
 */
 _arr32(char *a[]);

/**
  Constructor with/wo size and one element
  */
 _arr32(int, const char *);
 _arr32(const char *);
 
/**
  Complete constructor with one element.
  In order, the type, the starting capacity, and the string
  */
 _arr32(int, int, const char *);


/**
  Complete constructor with a text item.
  */
 _arr32(int, text &);

/**
  Returns the number of elements
  */
 inline int size() { return len; }

/**
  Returns true if the array is empty
  */
 inline bool empty() { return len == 0; }


 void resize(int);

 /**
   Clears the content.
   */
 void clear();


 int compare(char *,char *);
 bool same(char *,char *);

 void set();
 void set(int);
 void set(const _arr32 &);
 void set(_arr32 *);
 void set(char *[]);
 void set(char *);

 void operator=(const _arr32 &);
 void operator=(_arr32 *);
 void operator=(char *);

 void setchar(int, char *);
 
 inline void setAt(int i, bool b) { erase(i); }  // set nil
 void setAt(int, char *);
 inline void setAt(int i, text &t) { setAt(i, t.data); }

 /**
   Adds a second array having the same type.
   */
 void add(_arr32);

char * &operator[](int);
char * &operator[](const char *);
char * at(int);
char * at(text);
char * at(const char *);
 
 bool operator==(_arr32);
 bool operator==(bool);
 inline bool operator!=(_arr32 a) { return !operator==(a); }
 inline bool operator!=(bool b)   { return !operator==(b); }
 bool operator<(_arr32);
 bool operator>(_arr32);
 inline bool operator>=(_arr32 a) { return !operator<(a); }
 inline bool operator<=(_arr32 b) { return !operator>(b); }

 _arr32 operator+(_arr32);
 _arr32 operator+(char *);
 _arr32 operator+(text x);


 _arr32 operator-(_arr32);
 _arr32 operator-(char *);
 inline _arr32 operator-(text x)  { return operator-(x.data); }

 _arr32 operator&(_arr32);
 _arr32 operator&(char *);
 _arr32 operator&(text x);

 _arr32 operator|(_arr32);
 _arr32 operator|(char *);
 _arr32 operator|(text x);

// list of item not in both arrays
 _arr32 operator^(_arr32);

 void operator+=(_arr32);
 void operator+=(char *);
 inline void operator+=(text t)  { operator+=(t.data); }

 void operator-=(_arr32);
 void operator-=(char *);
 inline void operator-=(text t)  { operator-=(t.data); }

 void operator&=(_arr32);
 void operator&=(char *);
 inline void operator&=(text t)  { operator&=(t.data); }

 void operator|=(_arr32);
 void operator|=(char *);
 inline void operator|=(text t)  { operator|=(t.data); }

 void operator^=(_arr32);

 bool operator!();

 //===== util
 
 bool in(char *);
 signed int hasStrKey(const text &);
 signed int findKey(const text &);
 signed int findKey(int);
 signed int findValue(char *);

/**
  Finds the position of an element inside the array.
  Return the position or -1 if not found.
  */
 inline int find(char *x) { return findValue(x); }

/**
  Displays the array
  */
 void display();


 void display(int);


/**
  Joins the elements of the array into a string.
  Numbers are converted to strings of digits.
  */
 char *toStr();
 char *toString(const char *sep = ",");

 /**
  Joins the elements of the array into a text.
  */
 text toText();

 //===== iterator

/**
  Points out the indexer at the first element.
  */
char *begin();

/**
  Points out the indexer at the last element.
  */
char *end();

/**
  Restarts the iterator at first element.
  */
 void reset();
 bool outOfBound();
 bool bound();
 bool inBound();

/**
  Returns the current element and increments the indexer to the next position.
  */
char *inc();

/**
  Returns the current element and decrements the indexer to the previous position.
  */
char *dec();

/**
  Returns the current position of the iterator. Use index() instead.
  */
 int key();
 int key(int);

/**
  Returns the currently pointer out element.
  */
char *value();
char *value(int);
char *current();

/**
  Return the current location of the indexer.
  */
 int index();

 //===== stack

/**
  Pushes an string at the end of the array.
  */
 void push(char *);

/**
  Pushes a text at the end of the array.
  */
 inline void push(text t) { push(t.data); }

 void pushUnic(char *);

/**
  Returns the last element, and remove it.
  */
char *pop();

/**
  Returns the first element, and remove it.
  */
char *shift();

/**
  Inserts an object at beginning.
  */
 void unshift(char *);

/**
  Inserts a text at beginning.
  */
 inline void unshift(text &t) { unshift(t.data); }

 //===== ranges

/**
  Inserts an object at position.
  */
 void insert(int,char *);
 

/**
  Inserts an array at position. The elements of the array are inserted,
  not the array itself.
 */
 void insert(int, _arr32);


/**
  Removes a range of elements for the first to the last given positions,
  included. Use splicing instead.
  */
 void erase(int, int);

/**
  Removes the element at position.
  */

 void erase(int);
 _arr32 slice(int, int, int);
 inline _arr32 slice(int f) { return slice(f, len, 0); }

/**
  Extracts a sub-array from the array.
  Parameters are the position of the first and last element included.
  */
 inline _arr32 slice(int f, int l) { return slice(f, l, 1); }

 void splice(int);

/**
  Erases a range of elements.
  Parameters are the position of the first and last element included.
  */
 void splice(int, int);

 void splice(int, int, int);
 void splice(int, int, _arr32, int);
 void splice(int, int, char *, int);
 void splice(int, int, bool, int);

 //===== transform

 /**
   Remove all null records from the array
   */
 void pack();

 void swapData(int, int);
 void dynSort(int,int);

/**
  Reverses the content of the array
  */
 _arr32 reverse();

/**
  Sorts the array in ascending order.
  */
 void sort(int i = 0);
 void aSort(int i = 0);
 void kSort(int i = 0) { return; }

/**
  Removes doubloons
  */
  _arr32 unique();

/**
  This method returns the array itself and is useless.
  */
 _arr32 values();

/**
  Returns the indices as an array of integers.
  */
 _arr32 keys();

 //===== operations

/**
  Copies into another array.
  */
 void clone(_arr32 *);

/**
  Returns the smaller element.
  If elements are texts, they are considered to represent integers,
  and non-digits text have 0 value.
  */
 int min();

/**
  Returns the higher element.
  If elements are texts, they are considered to represent integers.
  */
 int max();

/**
  Returns the sum of elements.
  If elements are texts, they are considered to represent integers.
  */
 int sum();

/**
  Joins the elements into a text when they are string and thus, for an
  array of texts only.
  */
 text join(const char *s = ",");

 //===== I/O

/**
  Fills an array from a file (one element per line).
  */
 void load(const char *);

/**
 Store an array into file. The end of line code must be added
 to elements of the array. Elements are stored as text if they are texts,
 as binary if they are scalars. Object are not handled for now.
 */
 bool store(const char *);

 static void boundError(const text &);
};

std::ostream& operator<< (std::ostream &, _arr32 &);
//_arr32* operator+(_arr32*,char *);

#endif

